% Data from working paper by Thomas Coleman, Yuying Li, and Arun Verma
% "Reconstructing the Unknown Local Volatility Function"
% Uses analytical formula for mapping Local Volatility from Implied Vol
% Taken from "Local and Stochastic Volatility Models: An Investigation
% Into the Pricing of Exotic Equity Options" by Lisa Majmin (2005)
% M.Sc. dissertation, University of the Witwatersrand, Johannesburg, S.A.
% By Fabrice Douglas Rouah

clc; clear;

%% Call option data
% Spot price
S0 = 590;

% Implied volatility
IV = [...
190	168	133	113	102	97	120	142	169	200;
177	155	138	125	109	103	100	114	130	150;
172	157	144	133	118	104	100	101	108	124;
171	159	149	137	127	113	106	103	100	110;
171	159	150	138	128	115	107	103	99	108;
169	160	151	142	133	124	119	113	107	102;
169	161	153	145	137	130	126	119	115	111;
168	161	155	149	143	137	133	128	124	123;
168	162	157	152	148	143	139	135	130	128;
168	164	159	154	151	148	144	140	139	132]./1000;

IV=IV-ones(size(IV))*0.01;

% Strikes
K = [501.50 531.00 560.50 590.00 619.50 649.00 678.50 708.00 767.00 826.00];
NK = length(K);

% Maturities T
T = [0.175 0.425 0.695 0.94 1 1.5 2 3 4 5];
NT = length(T);

% Current time
T0 = 0;

% Risk free rate 
r = 0.06;

% Dividend yield 
q = 0.0262;

% Call prices (Kind of silly for this example we don't need it)
for i=1:NT
    for j=1:NK
        C(i,j) = BSPrice(S0, K(j), r, T(i), IV(i,j), q, 'C');
    end;
end;

%% Expansion of strike and maturity domain for interpolation grid
% Append increments of (ki) and (2*ki) for each strike
% Define strike increment ki to be 1, which can be changed
ki = 1;
KI = [];
for i=1:NK
    KI = [KI, K(i)-2*ki, K(i)-ki, K(i), K(i)+ki, K(i)+2*ki];
end

% Append increments of (ti) and (2*ti) for each maturity
% Define maturity increment ti to be 0.001, which can be changed.
ti = 0.001;
TI = [];
for i=1:NT
    TI = [TI, T(i)-2*ti, T(i)-ti, T(i), T(i)+ti, T(i)+2*ti];
end

% Number of strikes and maturities in interpolation grid
NKI = length(KI);
NTI = length(TI);


%% Interpolate implied volatility along entire interpolation grid
% This is used to obtain partial derivatives of Implied Vol with
% respect to T and K
IVI = interp2(K, T, IV, KI', TI, 'spline');

%% Plot the interpolated implied volatility surface
surf(KI,TI,IVI)
xlabel('Strike')
ylabel('Maturity')

%pause

%% Partial derivatives of Implied Volatility wrt Strike
for i=2:NTI-1
    for j=2:NKI-1
        Forward(i,j)  = (IVI(i,j+1) - IVI(i,j)  ) / (KI(j+1) - KI(j));
        Backward(i,j) = (IVI(i,j)   - IVI(i,j-1)) / (KI(j)   - KI(j-1));
        DerK(i,j)     = (Forward(i,j) + Backward(i,j))/2;
    end;
end

DerK = [DerK; zeros(1,NKI-1)];
DerK = [DerK zeros(NTI,1)];
dVdK  =     DerK([3:5:NTI], [3:5:NKI]);
dVdKF =  Forward([3:5:NTI], [3:5:NKI]);
dVdKB = Backward([3:5:NTI], [3:5:NKI]);
clear Forward Backward

%% Partial derivatives of Implied Volatility wrt Maturity
for j=2:NKI-1
    for i=2:NTI-1
        Forward(i,j)  = (IVI(i+1,j) - IVI(i,j)  ) / (TI(i+1) - TI(i));
        Backward(i,j) = (IVI(i,j)   - IVI(i-1,j)) / (TI(i)   - TI(i-1));
        DerT(i,j)     = (Forward(i,j) + Backward(i,j))/2;
    end
end

DerT = [DerT; zeros(1,NKI-1)];
DerT = [DerT zeros(NTI,1)];
dVdT =      DerT([3:5:NTI], [3:5:NKI]);
dVdTF =  Forward([3:5:NTI], [3:5:NKI]);
dVdTB = Backward([3:5:NTI], [3:5:NKI]);
clear Forward Backward

%% Second partial derivative of Volatility wrt Strikes
clear DerK2 Forward Backward

for i=3:NTI-2
    for j=3:NKI-2
        Forward(i,j)  = (DerK(i,j+1) - DerK(i,j)  ) / (KI(j+1) - KI(j));
        Backward(i,j) = (DerK(i,j)   - DerK(i,j-1)) / (KI(j)   - KI(j-1));
        DerK2(i,j)    = (Forward(i,j) + Backward(i,j))/2;
    end
end

DerK2 = [DerK2; zeros(2,NKI-2)];
DerK2 = [DerK2  zeros(NTI,2)];
dV2dK2  =    DerK2([3:5:NTI], [3:5:NKI]);
dV2dK2F =  Forward([3:5:NTI], [3:5:NKI]);
dV2dK2B = Backward([3:5:NTI], [3:5:NKI]);
clear Forward Backward

%% Local Variance using Dupire's equation and the analytical form
% that appears in Lisa Majmin's M.Sc. dissertation (2005)

for i=1:NT
    for j=1:NK
        Num = 2*IV(i,j)*dVdT(i,j)*T(i) + IV(i,j)^2 + ...
              2*IV(i,j)*(r-q)*K(j)*T(i)*dVdK(i,j);
         d1 = (log(S0/K(j)) + (r-q-IV(i,j)^2/2)*T(i)) / IV(i,j) / sqrt(T(i));
        Den = (1+ K(j)*d1*dVdK(i,j)*sqrt(T(i)))^2 + K(j)^2*T(i)*IV(i,j)*(dV2dK2B(i,j) - d1*dVdK(i,j)^2*sqrt(T(i)));
        LocalVar(i,j) = Num/Den;
    end
end

%% Get rid of negative values in Local Variance using linear interpolation
for i=1:NT;
    Y = LocalVar(i,:) 
    I = find(Y<=0);
    all = [1:length(Y)];
    new = setdiff(all,I);
    newY = Y(new);
    newX = K(new);
    xi = K(I);
    yi = interp1(newX,newY,xi,'linear', 'extrap');
    newX = [newX xi];
    newY = [newY yi];
    A = sortrows([newX' newY']);
    newX = A(:,1);
    newY = A(:,2);
    LocalVar(i,:) = newY' 
end

%% Obtain Local volatility as the square root of local variance
LV = sqrt(LocalVar);


%% Interpolate the Local volatility surface
incrementI = (T(NT) - T(1))/NT/10;
incrementJ = (K(NK) - K(1))/NK/10;
Ti = T(1):incrementI:T(NT);
Ki = K(1):incrementJ:K(NK);
LVI = interp2(K, T, LV, Ki', Ti, 'spline');

%% Plot the local volatility surface
% See Coleman article, Figures 6 and 7
surf(Ki,Ti,LVI)
xlabel('Strike')
ylabel('Maturity')
